import logging
import argparse
from abc import ABC, abstractmethod

"""
Core (abstract) Processing module for data access across a network

Defines the following methods:
    _defineArgs() (abstract): allows the overriding class to define a custom list of appropriate arguments for the data 
        request
    run(): Assumes a call is made from console which will utilize the _defineArgs function to define the relevant args
        for the data call in conjunction with argparse to determine the parameters provided by the user. Arguments are 
        then passed to the main function which maintains the logic to actually perform the data request
    main(): Validates the input arguments and passes them to the requestData() function.
    requestData() (abstract):  Allows the overriding class to define a specific method to call for data. This could be 
        sql, graphql, webrequest, etc. Expected to return a data set in an undefined form. 
    _validateArgs(): Anticipated to be used by the inherited class within the main function to validate that the args
        that are provided are appropriate per the definition in the _defineArgs() method
"""




class DataAccessExecutor(ABC):
    def __init__(self, label, loggingLvl = None):
        self.label = label
        self.requestArgs = {}
        self.loggingLvl = loggingLvl

    @abstractmethod
    def _defineInputArgs(self):
        pass

    def run(self):
        # Gather Arguments
        parser = argparse.ArgumentParser(description=self.label)
        arg_defs = self._defineInputArgs()

        for arg in arg_defs:
            if arg.get('required', False) == True:
                parser.add_argument('-' + str(arg['short']), '--' + str(arg['long']),
                                type=arg['type'],
                                help=arg['help'],
                                required=True)
            else:
                parser.add_argument('-' + str(arg['short']), '--' + str(arg['long']),
                                    default=arg['default'],
                                    type=arg['type'],
                                    help=arg['help'])

        args = parser.parse_args()
        return self.main(vars(args), context="CONSOLE")

    def main(self, inargs, context=None, **kwargs):
        if context is None:
            context = self.label

        validInputArgs = self._validateInputArgs(inargs, context)
        requestArgs = self._defineRequestArgs()

        return self.request(validInputArgs, requestArgs, context, **kwargs)

    @abstractmethod
    def request(self, inputargs, requestargs, context, **kwargs):
        pass

    @abstractmethod
    def _defineRequestArgs(self):
        pass

    def _validateInputArgs(self, inargs, context=None):
        inargs = inargs if inargs else {}
        logging.log(self.loggingLvl, f"[{context}]: Input Args for [{self.label}]: {inargs}")

        args = self._defineInputArgs()
        ret = {}
        for arg in args:
            """
            New dictionary will have data in the form: 
             {   key == the parameter name defined in the calling class
                 value == value provided by the user, method or default if not a required param}
            """
            if arg.get('required', False) == False:
                ret[arg['long']] = inargs.get(arg['long'], arg['default'])
            else:
                if arg['long'] in inargs:
                    ret[arg['long']] = inargs[arg['long']]
                else:
                    raise Exception(f"value {arg['long']} was not provided")

        logging.log(self.loggingLvl, f"[{context}]: Validated Args for [{self.label}]: {ret}")

        return ret




from cooptools.coopEnum import CoopEnum
from dataclasses import dataclass

class DataBaseType(CoopEnum):
    MYSQL = 'mssql'
    SQLSERVER = 'mssql'

class DataBaseConnector(CoopEnum):
    PYODBC = 'pyodbc'

@dataclass
class DbConnectionArgs:
    db_type: DataBaseType
    db_connector: DataBaseConnector
    server_name: str
    db_name: str
    port: int = None
    trusted_connection: bool = True
    driver: str = "ODBC+Driver+17+for+SQL+Server"
    user: str = None
    pw: str = None

    def sqlalchemy_db_connection_string(self, echo: bool = False):

        up_txt = ""
        if not self.trusted_connection:
            if self.user is None or self.pw is None:
                raise ValueError("User and password cannot be None when it is not a trusted connection")
            up_txt = f"{self.user}:{self.pw}"

        trusted_txt = "?TrustedConnection=yes" if self.trusted_connection else ""

        server_txt = f"{self.server_name}:{self.port}" if self.port else f"{self.server_name}"

        driver = self.driver.replace(" ", "+")

        conn_str = f"{self.db_type.value}+{self.db_connector.value}://{up_txt}@{server_txt}/{self.db_name}?driver={driver}{trusted_txt}"

        if echo:
            print(conn_str)

        return conn_str


import os
import sqlalchemy as db
from sqlalchemy.orm import Session
from typing import Callable
from coopio.sql.dbconnection_args import *
import logging
from coopio.sql.SqlReturnObj import SqlReturnObj
import time
import pandas as pd
import coopio.sql.SQLLogging as slog
import dataclasses
import pyodbc

def get_user_login():
    user = os.getlogin()
    domain = os.environ['userdomain']

    return fr"{domain}\{user}"

def create_a_db_connection_string(server_name: str,
                                  db_name: str,
                                  trusted_connection: bool = True,
                                  driver: str = "ODBC+Driver+17+for+SQL+Server",

                                  user: str = None,
                                  pw: str = None):

    trusted_txt = 'yes' if trusted_connection else 'no'
    up_txt = ""

    if not trusted_connection:
        if user is None or pw is None:
            raise ValueError("User and password cannot be None when it is not a trusted connection")

        up_txt = f"&User ID={user}&Password={pw}"

    driver = driver.replace(" ", "+")



    return f'mssql+pyodbc://@{server_name}/{db_name}?trusted_connection={trusted_txt}&driver={driver}{up_txt}'


def get_sql_connection(db_connection_args: DbConnectionArgs,
                       echo: bool = False,
                       autocommit: bool = True,
                       context: str = None):
    # Update connection to newly created db
    connection_str = db_connection_args.sqlalchemy_db_connection_string()

    try:
        slog.logger.log(slog.QUERY_LEVEL_NUM, f"Establishing connection to SQL Server. Context [{context}] -- {connection_str}")

        conn = db.create_engine(connection_str,
                                connect_args={'autocommit': autocommit},
                                echo=echo)
        slog.logger.log(slog.QUERY_LEVEL_NUM, f"Connection established! Context: [{context}]")
        return conn

    except pyodbc.InterfaceError as e:
        error = f"Context [{context}]: {e}"
        slog.logger.error(error)
        raise Exception(error)
    except Exception as e:
        error = f"Context [{context}]: {e}"
        slog.logger.error(error)
        raise Exception(error)



def create_user_and_add_server_role(db_connection_args: DbConnectionArgs, windows_user: str):

    actual_args = dataclasses.replace(db_connection_args)
    actual_args.db_name = 'master'

    # get connection to master
    sqlcon = get_sql_connection(actual_args)

    # run sql
    with sqlcon.connect() as connection:
        sql = f"USE [master]"
        deb = connection.execute(sql)

        sql = f"\nif not exists(select * from sys.server_principals where name = '{windows_user}')" \
              f"\nBegin" \
              f"\n    CREATE LOGIN [{windows_user}] FROM WINDOWS WITH DEFAULT_DATABASE=[master], DEFAULT_LANGUAGE=[us_english]" \
              f"\nEND"
        deb = connection.execute(sql)

        sql = f"\nALTER SERVER ROLE [sysadmin] ADD MEMBER [{windows_user}]"
        deb = connection.execute(sql)


def get_session(db_connection_args: DbConnectionArgs):
    sqlcon = get_sql_connection(db_connection_args)

    return Session(sqlcon)

def check_for_db(db_connection_args: DbConnectionArgs):
    verify_db_sql = f"select * from sys.databases where name = '{db_connection_args.db_name}'"

    actual_args = dataclasses.replace(db_connection_args)
    actual_args.db_name = 'master'

    # get connection to master
    sqlcon = get_sql_connection(actual_args)

    # verify the db exists
    with sqlcon.connect() as connection:
        results = connection.execute(verify_db_sql)
        rows = results.fetchall()

    if len(rows) == 1:
        return True
    return False

def check_connections_to_db(db_connection_args: DbConnectionArgs, kill: bool = False):

    actual_args = dataclasses.replace(db_connection_args)
    actual_args.db_name = 'master'

    # get connection to master
    sqlcon = get_sql_connection(actual_args)

    with sqlcon.connect() as connection:
        sql = f"if object_id('tempdb..#TEMP') is not null drop table #TEMP;	"
        results = connection.execute(sql)
        sql =   f"\ncreate table #TEMP											"\
                f"\n(															"\
                f"\n	SPID int,												"\
                f"\n	Status nvarchar(max),									"\
                f"\n	Login nvarchar(max),									"\
                f"\n	HostName nvarchar(max),									"\
                f"\n	BlkBy nvarchar(max),									"\
                f"\n	DBName nvarchar(max),									"\
                f"\n	Command nvarchar(max),									"\
                f"\n	CPUTime int,											"\
                f"\n	DiskIO int,												"\
                f"\n	LastBatch nvarchar(max),								"\
                f"\n	ProgramName nvarchar(max),								"\
                f"\n	SPID2 int,												"\
                f"\n	REQUESTID int											"\
                f"\n);															"
        results = connection.execute(sql)
        sql =   f"\ninsert into #TEMP											"\
                f"\nexec sp_who2;												"
        results = connection.execute(sql)
        sql = f"\nselect * from #TEMP where DBName = '{db_connection_args.db_name}';     		"
        results = connection.execute(sql)
        rows = results.fetchall()


        if kill:
            for spid in rows:
                connection.execute(f"kill {spid[0]}")

    return results

def delete_db(db_connection_args: DbConnectionArgs):
    # delete db string
    drop_db_sql = f"DROP DATABASE IF EXISTS {db_connection_args.db_name};"

    actual_args = dataclasses.replace(db_connection_args)
    actual_args.db_name = 'master'

    # get connection to master
    sqlcon = get_sql_connection(actual_args)

    # check and kill open connections
    connections = check_connections_to_db(db_connection_args, kill=True)

    # drop db
    with sqlcon.connect() as connection:
        deb = connection.execute(drop_db_sql)

def create_db(db_connection_args: DbConnectionArgs, echo: bool = False):
    actual_args = dataclasses.replace(db_connection_args)
    actual_args.db_name = 'master'

    # get connection to master
    sqlcon = get_sql_connection(actual_args)

    create_db_sql = f"IF DB_ID('{db_connection_args.db_name}') IS NULL" \
                     f"\nBEGIN" \
                     f"\nCREATE DATABASE {db_connection_args.db_name};" \
                     f"\nEND\n\n"

    # run sql
    with sqlcon.connect() as connection:
        deb = connection.execute(create_db_sql)

def connect_to_db(db_connection_args: DbConnectionArgs, recreate_if_existing: bool, create_if_missing: bool=True, echo: bool = False):
    # check for db
    db_exists = check_for_db(db_connection_args)

    # recreate db per param
    if recreate_if_existing and db_exists:
        delete_db(db_connection_args)
        create_db(db_connection_args)

    # create db per param
    if create_if_missing and not db_exists:
        create_db(db_connection_args)

    # Update connection to newly created db
    return get_sql_connection(db_connection_args, echo=echo)

def _execute(connection, sql, title):
    deb = connection.execute(sql)
    data = deb.fetchall()
    myData = pd.DataFrame(data)

    # myData = pd.read_sql(sql, conn)
    slog.logger.query(f"Data Returned for {title}: \n\t{myData}"
                      f"\n\tusing the query:"
                      f"\n\t\"{sql}\"")
    return myData

def execute_sql(sql: str,
                db_connection_args: DbConnectionArgs,
                autocommit: bool = False,
                title: str = None,
                echo: bool = False,
                existing_conn = None) -> SqlReturnObj:

    start_time = time.perf_counter()
    title_txt = f"the {title}" if title is not None else "a "
    slog.logger.query(f"Running {title_txt} query\n")

    error = None
    myData = None
    try:
        if existing_conn:
            myData = _execute(existing_conn, sql, title)
            print("used existing")
        else:
            conn = get_sql_connection(db_connection_args, echo=echo, autocommit=autocommit)

            with conn.connect() as connection:
                myData = _execute(connection, sql, title)

    except Exception as e:
        slog.logger.error(f"Unable to return data from source: {e}, {type(e)}")
        error = e
        myData = None
    finally:
        end_time = time.perf_counter()
        delta_t = (end_time - start_time)
        slog.logger.query(f"Elapsed time: {delta_t} seconds")
        return SqlReturnObj(error=error, data=myData, timer=f"{delta_t:.4f} sec")




if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)


    db_connect_args = DbConnectionArgs(
        db_type=DataBaseType.SQLSERVER,
        db_connector=DataBaseConnector.PYODBC,
        server_name='172.21.3.151',
        db_name='MaestroAnalytics',
        trusted_connection=False,
        user='asapdb',
        pw="K943$bY2pr"
    )

    existing_conn = get_sql_connection(db_connect_args)
    with existing_conn.connect() as connection:
        ret = execute_sql("select top 1000 * from log.analytics_log", db_connect_args, existing_conn=connection)




from typing import List, Dict, Any
from coopio.IDataService import IDataService, T
import logging
from sqlalchemy.orm import Session
from sqlalchemy.inspection import inspect
import pandas as pd
from mapper.object_mapper import ObjectMapper
from coopio.sql.utils import *

class SqlDataService(IDataService):

    def __init__(self,
                 servername: str,
                 db_name: str,
                 base: db.orm.decl_api.DeclarativeMeta,
                 orm_obj_mapping_factory: ObjectMapper = None,
                 recreate_db_if_existing:bool=False,
                 create_if_missing: bool = True,
                 echo: bool = False):

        self.servername = servername
        self.db_name = db_name
        self.base = base
        self.orm_obj_mapping_factory = orm_obj_mapping_factory
        super().__init__()

        # create login with execution creds
        create_user_and_add_server_role(servername, get_user_login())

        # connect to db
        sqlcon = connect_to_db(servername, db_name, recreate_if_existing=recreate_db_if_existing, create_if_missing=create_if_missing, echo=echo)

        # Create defined tables
        with sqlcon.connect() as connection:
            self.base.metadata.create_all(bind=connection)
            connection.close()
        sqlcon.dispose()


    @staticmethod
    def _commit(engine, session):
        try:
            try:
                session.commit()
                return True
            except db.exc.SQLAlchemyError as e:
                logging.error(e)
                session.rollback()
                return False
        except:
            engine.dispose()
            return False

    def _input_type_to_mapping(self, obj_type: T):
        # no mapper, assume working on raw ORM objects
        if self.orm_obj_mapping_factory is None:
            return obj_type

        # object not mapped, will attempt to return the object entered
        mapping = self.orm_obj_mapping_factory.mappings.get(obj_type, None)
        if mapping is None:
            return obj_type

        # has a mapping, return the mapped item type (choose first mapping in dict)
        return list(mapping.keys())[0]

    def _try_bulk_save_fail_to_individual_with_rollback(self, session, objs):
        try:
            session.add_all(objs)
        except Exception as e:
            issues = []
            for obj in objs:
                try:
                    session.add(obj)
                except Exception as iner:
                    issues.append((obj, iner))

            issues_iter = iter([f"{x[0]} -- {x[1]}" for x in issues])
            issues_txt = ("\n").join(issues_iter)

            session.rollback()
            raise Exception(f"Unable to bulk load the objects. Exception {e} was raised when bulk inserting. The issues were on"
                            f"{issues_txt}")

    def add_or_update(self, obj_type: T, objs: List[T], ret_as_orm: bool = False, try_map: bool = True, **kwargs) -> List[T]:

        entries_to_update = []
        entries_to_put = []

        copy_objs = objs.copy()

        if try_map:
            orm_type = self._input_type_to_mapping(obj_type)
        else:
            orm_type = obj_type

        sqlcon = get_sql_connection(self.servername, self.db_name)
        with Session(sqlcon) as session:
            try:
                # Find all objects that needs to be updated
                primary_key = inspect(orm_type).primary_key[0].name # https://stackoverflow.com/questions/6745189/how-do-i-get-the-name-of-an-sqlalchemy-objects-primary-key
            except Exception as e:
                raise Exception(f"Unable to inpsect the primary key values of type {orm_type}")

            primary_identifiers = [getattr(obj, primary_key) for obj in copy_objs]

            for each in self._try_retrieve_objs_of_type(session, orm_type, primary_key, primary_identifiers):
                # obj = objs.pop(getattr(each, primary_key))
                index = next(idx for idx in range(len(copy_objs)) if getattr(copy_objs[idx], primary_key) == getattr(each, primary_key))
                obj = copy_objs.pop(index)
                entries_to_update.append(obj)

            # Bulk mappings for everything that needs to be inserted
            for obj in copy_objs:
                entries_to_put.append(obj)

            # bulk save
            if self.orm_obj_mapping_factory is not None and try_map:
                puts = [self.orm_obj_mapping_factory.map(x) for x in entries_to_put]
                self._try_bulk_save_fail_to_individual_with_rollback(session, puts)
            else:
                self._try_bulk_save_fail_to_individual_with_rollback(session, entries_to_put)

            # merge objects that were already in db
            if self.orm_obj_mapping_factory is not None and try_map:
                updts = [self.orm_obj_mapping_factory.map(x) for x in entries_to_update]
            else:
                updts = entries_to_update
            for obj in updts:
                session.merge(obj)

            # commit
            if not (self._commit(sqlcon, session)):
                if kwargs.get('allow_partial', True) and len(copy_objs) > 1:
                    for obj in copy_objs:
                        self.add_or_update(obj_type, [obj], allow_partial=False)
                else:
                    obj_txt = self._lots_of_objects_to_string(copy_objs)
                    raise Exception(f"Unable to commit the add_or_update operation for objects {obj_type}"
                                    f"\n{obj_txt}")

            # return
            return self.retrieve_objs(obj_type, [getattr(obj, primary_key) for obj in objs], ret_as_orm=ret_as_orm, try_map=try_map)

    @staticmethod
    def _lots_of_objects_to_string(objs, n_objs_to_show_start = 10, n_objs_to_show_end = 10):

        if len(objs) <= n_objs_to_show_start + n_objs_to_show_end:
            obj_txt = "\n".join(iter([str(x) for x in objs]))
        else:
            obj_txt = "\n".join(iter([str(x) for x in objs[:n_objs_to_show_start]]))
            obj_txt += f"\n...{len(objs) - n_objs_to_show_start - n_objs_to_show_end} objects omitted...\n"
            obj_txt += "\n".join(iter([str(x) for x in objs[-n_objs_to_show_end:]]))

        return obj_txt

    @staticmethod
    def _try_retrieve_objs_of_type(session, orm_type: T, key: str, ids: List[Any] = None):
        def batch_them(iterable, n=1):
            l = len(iterable)
            for ndx in range(0, l, n):
                yield iterable[ndx:min(ndx + n, l)]

        try:
            if ids is not None:
                orm_results = []
                # must batch since there is a max length of 1000 on the .in_ function

                for batch in batch_them(ids, 1000):
                    batch_results = session.query(orm_type).filter(getattr(orm_type, key).in_(batch)).all()
                    orm_results += batch_results
            else:
                orm_results = session.query(orm_type).all()

            return orm_results
        except Exception as e:
            raise Exception(f"Error querying {orm_type}"
                            f"\nINNER: {type(e)}"
                            f"\n{e}")


    def retrieve_objs(self, obj_type: T, ids: List[Any] = None, ret_as_orm: bool = False, try_map: bool = True, **kwargs) -> List[T]:

        if try_map:
            orm_type = self._input_type_to_mapping(obj_type)
        else:
            orm_type = obj_type

        sqlcon = get_sql_connection(self.servername, self.db_name)
        with Session(sqlcon) as session:
            primary_key = inspect(orm_type).primary_key[0].name # https://stackoverflow.com/questions/6745189/how-do-i-get-the-name-of-an-sqlalchemy-objects-primary-key

            #query
            orm_results = self._try_retrieve_objs_of_type(session, orm_type, primary_key, ids)

            # map and return
            if self.orm_obj_mapping_factory is not None and not ret_as_orm and try_map:
                ret = [self.orm_obj_mapping_factory.map(x) for x in orm_results]
            else:
                ret = orm_results

            return ret

    def delete(self, obj_type: T, ids: List[Any] = None) -> Dict[str, bool]:
        sqlcon = get_sql_connection(self.servername, self.db_name)
        with Session(sqlcon) as session:
            objs = self.retrieve_objs(obj_type, ids, ret_as_orm=True)

            [session.delete(obj) for obj in objs]

            for obj in objs:
                session.delete(obj)

            # commit
            if not (self._commit(sqlcon, session)):
                raise Exception(f"Unable to commit the delete operation for objects {obj_type}")

            return {id: True for id in ids} if ids else {}

    def delete_db(self):
        delete_db(self.servername, self.db_name)

    def retrieve_as_df(self, obj_type: T, ids: List[Any] = None) -> pd.DataFrame:
        objs = self.retrieve_objs(obj_type, ids)
        df = pd.DataFrame([vars(x) for x in objs])
        return df

    def translate_from_data_rows(self, obj_type: T, df: pd.DataFrame) -> List[T]:
        raise NotImplementedError()



from unittest import TestCase
from unittest.mock import Mock, MagicMock
from tests.dummies import Dummy, DummyDataProcessor, init_sql_dataservice, ORM_Dummy, CsvDataService_Dummy
from coopio.sql.SQLDataService import delete_db
import os


class Test_IDataProcessor(TestCase):
    # Create DataService
    servername = r'localhost\SQLEXPRESS'
    db_name = 'coopio_testdb'

    def tearDown(self) -> None:
        delete_db(self.servername, self.db_name)

    def test_add(self):
        '''Arrange'''
        mock_primary = Mock()
        mock_secondary = Mock()

        data_processor = DummyDataProcessor(mock_primary, [mock_secondary])

        dummies = []
        for ii in range(0, 5):
            dummy = Dummy(
                id=ii,
                a=ii,
                b=2,
                c=3
            )
            dummies.append(dummy)

        '''Act'''
        data_processor.add_or_update(obj_type=Dummy, objs=dummies)

        '''Assert'''
        mock_primary.add_or_update.assert_called_once_with(obj_type=Dummy, objs=dummies)
        mock_secondary.add_or_update.assert_called_once_with(obj_type=Dummy, objs=dummies)

    def test_retrieve(self):
        '''Arrange'''
        mock_primary = Mock()
        mock_secondary = Mock()

        data_processor = DummyDataProcessor(mock_primary, [mock_secondary])

        '''Act'''
        ret_ids = ["1", "2"]
        data_processor.retrieve_objs(Dummy, ids=ret_ids)

        '''Assert'''
        mock_primary.retrieve_objs.assert_called_once_with(obj_type=Dummy, ids=ret_ids)
        mock_secondary.retrieve_objs.assert_not_called()

    def test_delete(self):
        '''Arrange'''
        mock_primary = Mock()
        mock_secondary = Mock()

        mock_primary.delete = MagicMock(return_value={})
        mock_secondary.delete = MagicMock(return_value={})

        data_processor = DummyDataProcessor(mock_primary, [mock_secondary])
        dummies = []
        for ii in range(0, 5):
            dummy = Dummy(
                id=ii,
                a=ii,
                b=2,
                c=3
            )
            dummies.append(dummy)

        '''Act'''
        data_processor.delete(obj_type=Dummy, objs=dummies)

        '''Assert'''
        mock_primary.delete.assert_called_once_with(obj_type=Dummy, ids=[str(vars(dummy)[data_processor.obj_identifier]) for dummy in dummies])
        mock_secondary.delete.assert_called_once_with(obj_type=Dummy, ids=[str(vars(dummy)[data_processor.obj_identifier])  for dummy in dummies])


    def test_overwrite_peripheral_with_primary(self):
        '''Arrange'''
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '\\testData\\utdummies.csv'
        file_path = dir + data_file
        primary = CsvDataService_Dummy(file_path)
        secondary = init_sql_dataservice(self.servername, self.db_name)

        data_processor = DummyDataProcessor(primary, [secondary])
        dummies = []
        for ii in range(0, 5):
            dummy = Dummy(
                id=ii,
                a=ii,
                b=2,
                c=3
            )
            dummies.append(dummy)

        '''Act'''
        data_processor.add_or_update(obj_type=Dummy, objs=dummies)
        secondary.delete(obj_type=Dummy, ids=[x.id for x in dummies])

        secondary_objects = secondary.retrieve_objs(Dummy)
        self.assertEqual(len(secondary_objects), 0)

        data_processor.overwrite_peripherals_with_primary(Dummy)

        '''Assert'''
        primary_objects = primary.retrieve_objs(Dummy)
        secondary_objects = secondary.retrieve_objs(Dummy)

        self.assertEqual(len(primary_objects), len(secondary_objects))
        self.assertNotEqual(len(primary_objects), 0)





import pyodbc
import logging

# SQL Server Native Client 11.0
# SQL Server Native Client RDA 11.0
# ODBC Client 13 for SQL Server
# ODBC Client 17 for SQL Server
# SQL Server

def connect(connectionString:str,
            loggingLevel=logging.DEBUG,
            context=None,
            login_timeout=0,
            autocommit: bool = False):
    try:
        logging.log(loggingLevel, f"Establishing connection to SQL Server. Context [{context}] -- {connectionString}")
        conn = pyodbc.connect(connectionString, timeout=login_timeout, autocommit=autocommit)

        logging.log(loggingLevel, f"Connection established! Context: [{context}]")
        return conn
    except pyodbc.InterfaceError as e:
        error = f"Context [{context}]: {e}"
        logging.error(error)
        raise Exception(error)
    except Exception as e:
        error = f"Context [{context}]: {e}"
        logging.error(error)
        raise Exception(error)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format='%(name)s -- %(asctime)s : [%(levelname)s] %(message)s (%(filename)s lineno: %(lineno)d)')
    # sql_db_connection = "Driver={SQL Server Native Client 11.0};Server=192.168.11.4;Database=MaestroAnalytics;Trusted_Connection=no;UID=asapdb;PWD=asap;"
    sql_db_connection = r'Driver={SQL Server};Server=localhost\SQLEXPRESS;Database=master;Trusted_Connection=yes;'
    conn = connect(connectionString=sql_db_connection, loggingLevel=logging.DEBUG, context="TEST")


from unittest import TestCase
import os
from os import path
import pandas as pd
import time
from tests.dummies import Dummy, CsvDataService_Dummy

class Test_ICsvDataService(TestCase):

    def test_addobj(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            dummy = Dummy(
                id=0,
                a=1,
                b=2,
                c=3
            )

            ''' Act '''
            data_service.add_or_update(Dummy, objs=[dummy])

            ''' Assert '''
            df = pd.read_csv(file_path)
            self.assertEqual(df.shape[0], 1, f"{df.shape[0]} DNE 1")

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)

    def test_updateobj(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            dummy = Dummy(
                id=0,
                a=1,
                b=2,
                c=3
            )

            ''' Act '''
            data_service.add_or_update(Dummy, objs=[dummy])
            new_val = 100
            dummy.c = new_val
            data_service.add_or_update(Dummy, objs=[dummy])

            ''' Assert '''
            df = pd.read_csv(file_path)
            self.assertEqual(df.shape[0], 1, f"{df.shape[0]} DNE 1")
            self.assertEqual(df.at[0, 'c'], new_val,
                             f"{df.at[0, 'c']} DNE {new_val}")

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)

    def test_retrieve_obj_data__valid_values(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            identifier = 'a'
            dummies = []
            for ii in range(0, 5):
                dummies.append(Dummy(
                    id=0,
                    a=ii,
                    b=2,
                    c=3
                ))
            data_service.add_or_update(Dummy, objs=dummies)

            ''' Act '''
            ret_ids = ["1", "2", "3"]
            objs = data_service.retrieve_objs(Dummy, ids=ret_ids)

            ''' Assert '''
            self.assertIsNotNone(objs, "Expected real values")
            self.assertEqual(type(objs), list,
                             f"Return should be of type: {list}, but type {type(objs)} was returned")
            self.assertEqual(len(objs), 3, f"Wrong number of objs returned: {len(objs)} DNE 3")
            [self.assertIn(str(vars(obj)[identifier]), ret_ids) for obj in objs]

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)

    def test_retrieve_equity_data__invalid_values(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            identifier = 'a'

            for ii in range(0, 5):
                dummy = Dummy(
                    id=ii,
                    a=ii,
                    b=2,
                    c=3
                )
                data_service.add_or_update(Dummy, objs=[dummy])

            ''' Act '''
            ret_ids = ["a", "b", "c"]
            objs = data_service.retrieve_objs(Dummy, ids=ret_ids)

            ''' Assert '''
            self.assertIsNotNone(objs, "Expected real values for equities")
            self.assertEqual(type(objs), list,
                             f"Return should be of type: {list}, but type {type(objs)} was returned")
            self.assertEqual(len(objs), 0, f"Wrong number of equities returned: {len(objs)} DNE 0")
            [self.assertIn(str(vars(obj)[identifier]), ret_ids) for obj in objs]

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)

    def test_delete(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            identifier = 'a'

            dummies = []
            for ii in range(0, 5):
                dummy = Dummy(
                    id=ii,
                    a=ii,
                    b=2,
                    c=3
                )
                dummies.append(dummy)
            data_service.add_or_update(Dummy,  objs=dummies)

            ''' Act '''
            ret_ids = ["1", "2", "3"]
            successful_deletes = data_service.delete(obj_type=Dummy, ids=ret_ids)

            ''' Assert '''
            df = pd.read_csv(file_path)
            self.assertEqual(df.shape[0], 2, f"{df.shape[0]} DNE 2")
            [self.assertNotIn(row[identifier], ret_ids) for i, row in df.iterrows()]

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)

    def test_add_or_update__bulk_performance(self):
        dir = os.path.dirname(os.path.abspath(__file__))
        data_file = '/testData/utdummies.csv'
        file_path = dir + data_file

        try:
            ''' Arrange'''
            data_service = CsvDataService_Dummy(data_file_path=file_path)
            identifier = 'a'

            dummies = []
            for ii in range(0, 10000):
                dummies.append(Dummy(
                    id=ii,
                    a=ii,
                    b=2,
                    c=3
                ))

            stored = data_service.add_or_update(Dummy, objs=dummies)
            for obj in stored:
                obj.c = 100

            ''' Act '''
            time_start = time.perf_counter()
            updated_objs = data_service.add_or_update(Dummy, objs=stored)
            elapsed = time.perf_counter() - time_start


            ''' Assert '''
            self.assertLess(elapsed, 3)
            [self.assertEqual(str(vars(obj)['c']), '100') for obj in updated_objs]

        finally:
            ''' Cleanup '''
            if path.exists(file_path):
                os.remove(file_path)


import logging

QUERY_LEVEL_NUM = logging.DEBUG + 5
QUERY = "QUERY"
logger = logging.getLogger('COOPIO_SQL')
# logging.addLevelName(QUERY_LEVEL_NUM, QUERY)
#
# def query(self, message, *args, **kws):
#     if self.isEnabledFor(QUERY_LEVEL_NUM):
#         # Yes, logger takes its '*args' as 'args'.
#         self._log(QUERY_LEVEL_NUM, message, args, **kws)
# logging.Logger.query = query


def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    # >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    # >>> logging.getLogger(__name__).setLevel("TRACE")
    # >>> logging.getLogger(__name__).trace('that worked')
    # >>> logging.trace('so did this')
    # >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
       raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
       raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
       raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)

def init_logging_config():
    addLoggingLevel(QUERY, QUERY_LEVEL_NUM)

init_logging_config()


class SqlReturnObj():
    def __init__(self,
                 data=None,
                 return_type=None,
                 callback=None,
                 error=None,
                 timer=None):
        self.data = data
        self.return_type = return_type
        self.callback = callback
        self.error = error
        self.timer = timer
        self.n_rows = len(self.data) if self.data is not None else 0


    def as_dict(self):
        return {"data": self.data,
                "return_type": self.return_type,
                "callback": self.callback,
                "error": self.error,
                "timer": self.timer,
                "n_rows": self.n_rows}



import sqlalchemy as db
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import UUIDType
from uuid import uuid4
import random as rnd
import pandas as pd
import logging
import datetime
from coopio.sql.SQLDataService import SqlDataService, get_sql_connection, delete_db
from mapper.object_mapper import ObjectMapper

import unittest


letters = 'abcdefghijklmnopqrstuvwxyz'
# logging.basicConfig()
# logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

def pretty_print_dataframe(df: pd.DataFrame,
                           title: str = None,
                           display_width: int = 2000,
                           display_max_columns: int = 2000,
                           display_max_rows: int = 500,
                           float_format: str = '%.3f',
                           ):
    if title:
        print(title)

    with pd.option_context('display.max_rows', display_max_rows,
                           'display.max_columns', display_max_columns,
                           'display.width', display_width,
                           'display.float_format', lambda x: float_format % x):
        print(f"{df}\n")

Base = declarative_base()

class ORM_Account(Base):
    __tablename__ = "accounts"
    id = db.Column(UUIDType, primary_key=True)
    name = db.Column(db.String)
    active = db.Column(db.Boolean)

class Account:
    def __init__(self,
                 id,
                 name,
                 active):
        self.id = id
        self.name = name
        self.active = active



class Test_SQLDataService(unittest.TestCase):
    servername = r'localhost\SQLEXPRESS'
    db_name = 'unittestdbforSQLDataService'

    def _instantiate_ds(self, orm_obj_mapping_factory = None) -> SqlDataService:
        # Create DataService
        return SqlDataService(self.servername, self.db_name, Base, orm_obj_mapping_factory, recreate_db_if_existing=False, create_if_missing=True)

    def _addORMAccounts(self, ds, n):
        accounts = [ORM_Account(name=''.join(rnd.choices(letters, k=n)),
                            id=uuid4(),
                            active=rnd.choice([True, False])) for ii in range(5)]
        ds.add_or_update(ORM_Account, accounts)

    def _addAccounts(self, ds, n):
        accounts = [Account(name=''.join(rnd.choices(letters, k=n)),
                            id=uuid4(),
                            active=rnd.choice([True, False])) for ii in range(5)]
        ds.add_or_update(Account, accounts)

    def tearDown(self):
        delete_db(self.servername, self.db_name)

    def test_instantiatedb(self):
        ######### ARRANGE ############
        verify_db_sql = f"select * from sys.databases where name = '{self.db_name}'"

        # get a connection
        sqlcon = get_sql_connection(self.servername, "master")

        ########## ACT ###############
        ds = self._instantiate_ds()
        # verify the db exists
        with sqlcon.connect() as connection:
            results = connection.execute(verify_db_sql)
            rows = results.fetchall()

        ######### ASSERT ############
        self.assertEqual(len(rows), 1, f"DB {self.db_name} was not created successfully")

    def test_addobjects(self):
        ######### ARRANGE ############
        n_accounts = 5

        ########## ACT ###############
        ds = self._instantiate_ds()
        self._addORMAccounts(ds, n_accounts)

        # retrieve accounts
        accounts = ds.retrieve_objs(ORM_Account)

        ######### ASSERT ############
        self.assertEqual(len(accounts), n_accounts)
        self.assertEqual(type(accounts[0]), ORM_Account)

    def test_addobjectswithORMMapper(self):
        ######### ARRANGE ############
        n_accounts = 5
        orm_obj_mapping_factory = ObjectMapper()
        orm_obj_mapping_factory.create_map(ORM_Account, Account, {'id': lambda x: x.id})
        orm_obj_mapping_factory.create_map(Account, ORM_Account, {'id': lambda x: x.id})

        ########## ACT ###############
        ds = self._instantiate_ds(orm_obj_mapping_factory=orm_obj_mapping_factory)
        self._addAccounts(ds, n_accounts)

        # retrieve accounts
        accounts = ds.retrieve_objs(Account)

        ######### ASSERT ############
        self.assertEqual(len(accounts), n_accounts)
        self.assertEqual(type(accounts[0]), Account)


if __name__ == "__main__":
    unittest.main()